from shapely.geometry import LineString, MultiPolygon, Polygon
from shapely.ops import split

def intersects_antimeridian(polygon: Polygon, lon_threshold: int=100) -> bool:
    """
    Check if a polygon intersects the antimeridian (longitude 180/-180).
    This is done by checking if the polygon crosses the antimeridian by looking for a large jump in
    longitude between consecutive points. This is used for a grid of cells which
    are not expected to extend from -lon_threshold to lon_threshold, meaning that
    if this is the case, the polygon crosses the antimeridian and should be split
    accordingly.
    """
    coords = list(polygon.exterior.coords)
    crosses_antimeridian = any(
        (x1 > lon_threshold and x2 < -lon_threshold)
        or (x1 < -lon_threshold and x2 > lon_threshold)
        for (x1, y1), (x2, y2) in zip(coords, coords[1:] + [coords[0]])
    )
    return crosses_antimeridian


def _fix_antimeridian_split(polygon: Polygon) -> MultiPolygon:
    """
    Split a polygon that crosses the antimeridian (longitude 180/-180) into two parts.
    This is done by adding a line that crosses the antimeridian and using it to split the polygon.
    """

    eps = 0
    splitter = LineString([(-180, 90), (-180, -90)])
    coords = list(polygon.exterior.coords)

    # Normalize coordinates to split across the antimeridian
    normalized_coords = [(x - 360 if x > 0 else x, y) for x, y in coords]
    normalized_polygon = Polygon(normalized_coords)

    # Split the polygon using the splitter line
    split_result = split(normalized_polygon, splitter)

    # # Filter out non-polygon geometries (like LinearRings or Points that might be created)
    # split_polygons = [geom for geom in split_result if isinstance(geom, Polygon)]

    # Normalize the polygons to have their coordinates within the range (-180, 180)
    fixed_polygons = []
    for poly in split_result.geoms:
        new_coords = []
        if poly.centroid.x < -180:
            new_coords = [(x + 360, y) for x, y in poly.exterior.coords]
            new_coords = [
                (x + eps, y) if x == -180 else (x, y) for x, y in new_coords
            ]
            new_coords = [
                (x - eps, y) if x == 180 else (x, y) for x, y in new_coords
            ]
            fixed_polygons.append(Polygon(new_coords))
        else:
            new_coords = [(x, y) for x, y in poly.exterior.coords]
            new_coords = [
                (x + eps, y) if x == -180 else (x, y) for x, y in new_coords
            ]
            new_coords = [
                (x - eps, y) if x == 180 else (x, y) for x, y in new_coords
            ]
            fixed_polygons.append(Polygon(new_coords))
            # fixed_polygons.append(poly)

    # Return the resulting polygons as a MultiPolygon
    new_polygon = MultiPolygon(fixed_polygons)

    return new_polygon


def fix_antimeridian_split(polygon: Polygon, lon_threshold: int=100) -> Polygon | MultiPolygon:
    # A line that we will use to split the polygon at the antimeridian
    if intersects_antimeridian(polygon, lon_threshold):
        return _fix_antimeridian_split(polygon)
    else:
        return polygon  # Return the original polygon if it doesn't cross the antimeridian